#include "main.h"
#include <util/delay.h>
#include <avr/interrupt.h>
#include <ctype.h>

#include "MOV.h"
#include "Logic.h"
#include "Aritm.h"
#include "IncDec.h"
#include "Jump.h"
#include "Rotate.h"
#include "CallRet.h"
#include "Halt.h"
#include "InOut.h"
#include "monitor.h"	// monitor via RS232
#include "SDcard.h"		// SD card functions

// global vars
volatile uint8_t Areg,Breg,Creg,Dreg,Ereg,Hreg,Lreg,Mem;	// 8008 registers and M
volatile uint8_t  Sflag,Zflag,Pflag,Cflag;			// 8008 status flags
volatile uint16_t PC;	// program counter
volatile int8_t SPtr;	// Stack pointer
volatile uint8_t scanKeyb = 0;	// Keyboard scan mode, initiated by OUT 9, 254
volatile uint16_t Stack[STACKSIZE];	// i8008 stack, comprises 7 registers
static pt2func iSet[256];
const uint8_t KRom[] PROGMEM = {
 #include "K1000ROM.txt"	
};
const uint8_t K1000ZS[] PROGMEM = {		// K1003 character set
 #include "K1000ZS.txt"
};
volatile uint8_t KRam[6144];		// 2K Buffer + 4K K1000 Ram
volatile uint8_t dispBuffer[16];
volatile uint8_t dBptr = 0;
volatile uint16_t ColRow = 0;		// store keypress coordinates
volatile uint8_t sp_key = 0;		// special key: SER, MKL or MKS
volatile uint8_t keydown = 0;		// a key is pressed currently
volatile uint8_t dispDelay = 0;
volatile uint8_t ix = 0;			// pointer for RS232 incoming
volatile uint8_t ox = 0;			// pointer for RS232 reading
volatile uint8_t buf[256];	// RS232 receive buffer
volatile uint8_t StopTransmission = 0;	// flag for XOFF/XON protocol

// parity lookup table
const uint8_t parity[] PROGMEM = {
#include "parity.txt"
 };

ISR(USART0_RX_vect)
{
char c;
	c = UDR0;
	if (c == XOFF) {
		StopTransmission = 1;
		PORTD |= 0b00010000;	// BES on
	}					
	else if (c == XON) {
		StopTransmission = 0;
		PORTD &= 0b11101111;	// BES off
	}	
	else buf[ix++] = c;
}

ISR(TIMER1_OVF_vect)
{
static uint8_t r,c;
uint8_t x;
	TCNT1 = 65461;		// for ~ 60 Hz Display freq.
	PORTA = 0;			// erase previous number
	x = 1;
	if (dBptr) x=0;
	PORTB = x;			// 1 every 16. time, else 0
	PORTB |= 0b010;
	PORTB &= 0b000;		// impuls 0-1-0 for SCK on PB1
	PORTB |= 0b100;
	PORTB &= 0b000;		// impuls 0-1-0 for RCK on PB2, also switch off PB0
	if (sp_key) dispDelay = 7;		// keep Display alive while in Special mode
	if (dispDelay) {
		dispDelay--;	// count down to 0, then leave at 0
		PORTA = dispBuffer[dBptr]; // display current char, only if dispDelay > 0
	}	
	if ((x = PINC & 0b11111110)) {	// this is an assignment combined with a test!
		r = x; c = dBptr;
	}	
	dBptr++;
	if (dBptr > 15) {
		keydown = r;			// keydown is a flag, 0=FALSE, <any other value>=TRUE
		if ((r==2 && c>9) || (r==4 && c==1)) {		// MKL, MKS, MONITOR
			sp_key = c;			// set flag for special key request
		} 
		else if (sp_key && r==2 && c==7)	// 'PROGR EING' ends monitor
		{
			sp_key = 0;
		}
		else ColRow = (c<<8) | r;		
		r = c = dBptr = 0;
	}	
}  // END of Timer-Interrupt

int main()
{
uint8_t nxt;
	DDRA = 0b11111111;  // PA output display segments
	DDRB = 0b11111111;	// PB0..2 output shift registers, SPI pins set to OUT
	DDRC = 0b00000001;	// PC1..7 input key rows
	DDRD = 0b11111111;	// PD2..4,6,7 output status LEDs, PD5 unused, set to OUT

	PORTB = 0b00000000;		// set all OUT pins to 0 on port B
	PORTC = 0b00000001;		// deactivate PULL UPs for port C1..7 (external PULL DOWNs)

    // initialize Timer 1
	TIMSK1 |= _BV(TOIE1);	// activate interrupt on overflow for Timer 1
	TCCR1B |= (1<<CS12);	// prescaling factor 256
//	TCCR1B |= (1<<CS11);	// prescaling factor 8
//	TCCR1B |= (1<<CS10);	// no prescaling
	TCNT1 = 0xFFFF;			// FFFF for start
	
 	AssignInstructionSet();
	InitProcessor();
	uart_init();		// set up UART communication
	sei();				// allow interupts
	// i8008 processor loop:  
	while (1) {
  		nxt = memory(PC);	// next 8008 machine code
		Mem = memory(Lreg | Hreg<<8);	
		if (Cflag) Cflag = 0b00000001;	// IMPORTANT: make sure only Bit 0 of Carry flag used
		// cli();
  			(*iSet[nxt])();		// execute next i8008 machine instruction
		// sei();
		if (sp_key) SpecialKey();	// invoke SER,MKL or MKS action
	}	// end.while
}

// ************** END OF MAIN PROGRAM ****************

void SpecialKey(void)
{
	if (sp_key == 1) monitor();
	if (sp_key == 10) ReadCard();
	if (sp_key == 11) WriteCard();
	sp_key = 0;		// release special key mode
	_delay_ms(100);
	do {} while (ColRow);	// wait until key released

}

void InitProcessor()
{
	Areg = Breg = Creg = Dreg = Ereg = Hreg = Lreg = 0;
	PC = 0;
	SPtr = 0;
	Cflag = Zflag = Sflag = Pflag = 0;
}

void AssignInstructionSet()
{
	iSet[0] = &hlt;
	iSet[1] = &hlt;	
	iSet[2] = &rlc;
	iSet[3] = &rnc;
	iSet[4] = &adi;
	iSet[5] = &rst0;
	iSet[6] = &mvi_a;
	iSet[7] = &ret;
	iSet[8] = &inc_b;
	iSet[9] = &dec_b;

	iSet[10] = &rrc;
	iSet[11] = &rnz;
	iSet[12] = &adci;
	iSet[13] = &rst1;
	iSet[14] = &mvi_b;
	iSet[15] = &ret;
	iSet[16] = &inc_c;
	iSet[17] = &dec_c;
	iSet[18] = &ral;
	iSet[19] = &rp;

	iSet[20] = &sui;
	iSet[21] = &rst2;
	iSet[22] = &mvi_c;
	iSet[23] = &ret;
	iSet[24] = &inc_d;
	iSet[25] = &dec_d;
	iSet[26] = &rar;
	iSet[27] = &rpo;
	iSet[28] = &sbi;
	iSet[29] = &rst3;

	iSet[30] = &mvi_d;
	iSet[31] = &ret;
	iSet[32] = &inc_e;
	iSet[33] = &dec_e;
	iSet[34] = &hlt;
	iSet[35] = &rc;
	iSet[36] = &ani;
	iSet[37] = &rst4;
	iSet[38] = &mvi_e;
	iSet[39] = &ret;

	iSet[40] = &inc_h;
	iSet[41] = &dec_h;
	iSet[42] = &hlt;
	iSet[43] = &rz;
	iSet[44] = &xri;
	iSet[45] = &rst5;
	iSet[46] = &mvi_h;
	iSet[47] = &ret;
	iSet[48] = &inc_l;
	iSet[49] = &dec_l;

	iSet[50] = &hlt;
	iSet[51] = &rm;
	iSet[52] = &ori;
	iSet[53] = &rst6;
	iSet[54] = &mvi_l;
	iSet[55] = &ret;
	iSet[56] = &nop;
	iSet[57] = &nop;
	iSet[58] = &nop;
	iSet[59] = &rpe;

	iSet[60] = &cpi;
	iSet[61] = &rst7;
	iSet[62] = &mvi_m;
	iSet[63] = &ret;
	iSet[64] = &jnc;
	iSet[65] = &in_0;
	iSet[66] = &cnc;
	iSet[67] = &in_1;
	iSet[68] = &jmp;
	iSet[69] = &in_2;

	iSet[70] = &call;
	iSet[71] = &in_3;
	iSet[72] = &jnz;
	iSet[73] = &in_4;
	iSet[74] = &cnz;
	iSet[75] = &in_5;
	iSet[76] = &jmp;
	iSet[77] = &in_6;
	iSet[78] = &call;
	iSet[79] = &in_7;

	iSet[80] = &jp;
	iSet[81] = &out_0;
	iSet[82] = &cp;
	iSet[83] = &out_1;
	iSet[84] = &jmp;
	iSet[85] = &out_2;
	iSet[86] = &call;
	iSet[87] = &out_3;
	iSet[88] = &jpo;
	iSet[89] = &out_4;

	iSet[90] = &cpo;
	iSet[91] = &out_5;
	iSet[92] = &jmp;
	iSet[93] = &out_6;
	iSet[94] = &call;
	iSet[95] = &out_7;
	iSet[96] = &jc;
	iSet[97] = &out_8;
	iSet[98] = &cc;
	iSet[99] = &out_9;

	iSet[100] = &jmp;
	iSet[101] = &out_10;
	iSet[102] = &call;
	iSet[103] = &out_11;
	iSet[104] = &jz;
	iSet[105] = &out_12;
	iSet[106] = &cz;
	iSet[107] = &out_13;
	iSet[108] = &jmp;
	iSet[109] = &out_14;

	iSet[110] = &call;
	iSet[111] = &out_15;
	iSet[112] = &jm;
	iSet[113] = &out_16;
	iSet[114] = &cm;
	iSet[115] = &out_17;
	iSet[116] = &jmp;
	iSet[117] = &out_18;
	iSet[118] = &call;
	iSet[119] = &out_19;

	iSet[120] = &jpe;
	iSet[121] = &out_20;
	iSet[122] = &cpe;
	iSet[123] = &out_21;
	iSet[124] = &jmp;
	iSet[125] = &out_22;
	iSet[126] = &call;
	iSet[127] = &out_23;

	iSet[128] = &add_a;
	iSet[129] = &add_b;
	iSet[130] = &add_c;
	iSet[131] = &add_d;
	iSet[132] = &add_e;
	iSet[133] = &add_h;
	iSet[134] = &add_l;
	iSet[135] = &add_m;

	iSet[136] = &adc_a;
	iSet[137] = &adc_b;
	iSet[138] = &adc_c;
	iSet[139] = &adc_d;
	iSet[140] = &adc_e;
	iSet[141] = &adc_h;
	iSet[142] = &adc_l;
	iSet[143] = &adc_m;

	iSet[144] = &sub_a;
	iSet[145] = &sub_b;
	iSet[146] = &sub_c;
	iSet[147] = &sub_d;
	iSet[148] = &sub_e;
	iSet[149] = &sub_h;
	iSet[150] = &sub_l;
	iSet[151] = &sub_m;

	iSet[152] = &sbc_a;
	iSet[153] = &sbc_b;
	iSet[154] = &sbc_c;
	iSet[155] = &sbc_d;
	iSet[156] = &sbc_e;
	iSet[157] = &sbc_h;
	iSet[158] = &sbc_l;
	iSet[159] = &sbc_m;

	iSet[160] = &and_a;
	iSet[161] = &and_b;
	iSet[162] = &and_c;
	iSet[163] = &and_d;
	iSet[164] = &and_e;
	iSet[165] = &and_h;
	iSet[166] = &and_l;
	iSet[167] = &and_m;

	iSet[168] = &xor_a;
	iSet[169] = &xor_b;
	iSet[170] = &xor_c;
	iSet[171] = &xor_d;
	iSet[172] = &xor_e;
	iSet[173] = &xor_h;
	iSet[174] = &xor_l;
	iSet[175] = &xor_m;

	iSet[176] = &or_a;
	iSet[177] = &or_b;
	iSet[178] = &or_c;
	iSet[179] = &or_d;
	iSet[180] = &or_e;
	iSet[181] = &or_h;
	iSet[182] = &or_l;
	iSet[183] = &or_m;

	iSet[184] = &cmp_a;
	iSet[185] = &cmp_b;
	iSet[186] = &cmp_c;
	iSet[187] = &cmp_d;
	iSet[188] = &cmp_e;
	iSet[189] = &cmp_h;
	iSet[190] = &cmp_l;
	iSet[191] = &cmp_m;

	iSet[192] = &mov_aa;
	iSet[193] = &mov_ab;
	iSet[194] = &mov_ac;
	iSet[195] = &mov_ad;
	iSet[196] = &mov_ae;
	iSet[197] = &mov_ah;
	iSet[198] = &mov_al;
	iSet[199] = &mov_am;

	iSet[200] = &mov_ba;
	iSet[201] = &mov_bb;
	iSet[202] = &mov_bc;
	iSet[203] = &mov_bd;
	iSet[204] = &mov_be;
	iSet[205] = &mov_bh;
	iSet[206] = &mov_bl;
	iSet[207] = &mov_bm;

	iSet[208] = &mov_ca;
	iSet[209] = &mov_cb;
	iSet[210] = &mov_cc;
	iSet[211] = &mov_cd;
	iSet[212] = &mov_ce;
	iSet[213] = &mov_ch;
	iSet[214] = &mov_cl;
	iSet[215] = &mov_cm;

	iSet[216] = &mov_da;
	iSet[217] = &mov_db;
	iSet[218] = &mov_dc;
	iSet[219] = &mov_dd;
	iSet[220] = &mov_de;
	iSet[221] = &mov_dh;
	iSet[222] = &mov_dl;
	iSet[223] = &mov_dm;

	iSet[224] = &mov_ea;
	iSet[225] = &mov_eb;
	iSet[226] = &mov_ec;
	iSet[227] = &mov_ed;
	iSet[228] = &mov_ee;
	iSet[229] = &mov_eh;
	iSet[230] = &mov_el;
	iSet[231] = &mov_em;

	iSet[232] = &mov_ha;
	iSet[233] = &mov_hb;
	iSet[234] = &mov_hc;
	iSet[235] = &mov_hd;
	iSet[236] = &mov_he;
	iSet[237] = &mov_hh;
	iSet[238] = &mov_hl;
	iSet[239] = &mov_hm;

	iSet[240] = &mov_la;
	iSet[241] = &mov_lb;
	iSet[242] = &mov_lc;
	iSet[243] = &mov_ld;
	iSet[244] = &mov_le;
	iSet[245] = &mov_lh;
	iSet[246] = &mov_ll;
	iSet[247] = &mov_lm;

	iSet[248] = &mov_ma;
	iSet[249] = &mov_mb;
	iSet[250] = &mov_mc;
	iSet[251] = &mov_md;
	iSet[252] = &mov_me;
	iSet[253] = &mov_mh;
	iSet[254] = &mov_ml;
	iSet[255] = &hlt;
}

uint8_t memory(uint16_t PC)
{
	if(PC < 2048) return pgm_read_byte(&KRom[PC]);
	else if (PC < 6144) return KRam[PC];
	else return pgm_read_byte(&KRom[PC]); 
}
